@@ -48,59 +48,57 @@ class QgsExpressionFunction;
  * \ingroup core
  * \brief Class for parsing and evaluation of expressions (formerly called "search strings").
  * The expressions try to follow both syntax and semantics of SQL expressions.
-
-Usage:
-\code{.py}
-  exp = QgsExpression("gid*2 > 10 and type not in ('D','F')")
-  if exp.hasParserError():
-      # show error message with parserErrorString() and exit
-
-  result = exp.evaluate(feature, fields)
-  if exp.hasEvalError():
-      # show error message with evalErrorString()
-  else:
-      # examine the result
-\endcode
-
-\section value_logic Three Value Logic
-
-Similarly to SQL, this class supports three-value logic: true/false/unknown.
-Unknown value may be a result of operations with missing data (NULL). Please note
-that NULL is different value than zero or an empty string. For example
-3 > NULL returns unknown.
-
-There is no special (three-value) 'boolean' type: true/false is represented as
-1/0 integer, unknown value is represented the same way as NULL values: NULL QVariant.
-
-\section performance Performance
-
-For better performance with many evaluations you may first call prepare(fields) function
-to find out indices of columns and then repeatedly call evaluate(feature).
-
-\section type_conversion Type conversion
-
-Operators and functions that expect arguments to be of a particular
-type automatically convert the arguments to that type, e.g. sin('2.1') will convert
-the argument to a double, length(123) will first convert the number to a string.
-Explicit conversion can be achieved with to_int, to_real, to_string functions.
-If implicit or explicit conversion is invalid, the evaluation returns an error.
-Comparison operators do numeric comparison in case both operators are numeric (int/double)
-or they can be converted to numeric types.
-
-\section implicit_sharing Implicit sharing
-
-This class is implicitly shared, copying has a very low overhead.
-It is normally preferable to call `QgsExpression( otherExpression )` instead of
-`QgsExpression( otherExpression.expression() )`. A deep copy will only be made
-when prepare() is called. For usage this means mainly, that you should
-normally keep an unprepared master copy of a QgsExpression and whenever using it
-with a particular QgsFeatureIterator copy it just before and prepare it using the
-same context as the iterator.
-
-Implicit sharing was added in 2.14
-
+ *
+ * Usage:
+ * \code{.py}
+ *   exp = QgsExpression("gid*2 > 10 and type not in ('D','F')")
+ *   if exp.hasParserError():
+ *       # show error message with parserErrorString() and exit
+ *
+ *   result = exp.evaluate(feature, fields)
+ *   if exp.hasEvalError():
+ *       # show error message with evalErrorString()
+ *   else:
+ *       # examine the result
+ * \endcode
+ *
+ * \section value_logic Three Value Logic
+ *
+ * Similarly to SQL, this class supports three-value logic: true/false/unknown.
+ * Unknown value may be a result of operations with missing data (NULL). Please note
+ * that NULL is different value than zero or an empty string. For example
+ * 3 > NULL returns unknown.
+ *
+ * There is no special (three-value) 'boolean' type: true/false is represented as
+ * 1/0 integer, unknown value is represented the same way as NULL values: NULL QVariant.
+ *
+ * \section performance Performance
+ *
+ * For better performance with many evaluations you may first call prepare(fields) function
+ * to find out indices of columns and then repeatedly call evaluate(feature).
+ *
+ * \section type_conversion Type conversion
+ *
+ * Operators and functions that expect arguments to be of a particular
+ * type automatically convert the arguments to that type, e.g. sin('2.1') will convert
+ * the argument to a double, length(123) will first convert the number to a string.
+ * Explicit conversion can be achieved with to_int, to_real, to_string functions.
+ * If implicit or explicit conversion is invalid, the evaluation returns an error.
+ * Comparison operators do numeric comparison in case both operators are numeric (int/double)
+ * or they can be converted to numeric types.
+ *
+ * \section implicit_sharing Implicit sharing
+ *
+ * This class is implicitly shared, copying has a very low overhead.
+ * It is normally preferable to call `QgsExpression( otherExpression )` instead of
+ * `QgsExpression( otherExpression.expression() )`. A deep copy will only be made
+ * when prepare() is called. For usage this means mainly, that you should
+ * normally keep an unprepared master copy of a QgsExpression and whenever using it
+ * with a particular QgsFeatureIterator copy it just before and prepare it using the
+ * same context as the iterator.
+ *
+ * Implicit sharing was added in 2.14
 */
-
 class CORE_EXPORT QgsExpression
 {
     Q_DECLARE_TR_FUNCTIONS( QgsExpression )
@@ -240,6 +238,14 @@ class CORE_EXPORT QgsExpression
      * all attributes from the layer are required for evaluation of the expression.
      * QgsFeatureRequest::setSubsetOfAttributes automatically handles this case.
      *
+     * \warning If the expression has been prepared via a call to QgsExpression::prepare(),
+     * or a call to QgsExpressionNode::prepare() for a node has been made, then parts of
+     * the expression may have been determined to evaluate to a static pre-calculatable value.
+     * In this case the results will omit attribute indices which are used by these
+     * pre-calculated nodes, regardless of their actual referenced columns.
+     * If you are seeking to use these functions to introspect an expression you must
+     * take care to do this with an unprepared expression.
+     *
      * \see referencedAttributeIndexes()
      */
     QSet<QString> referencedColumns() const;
@@ -249,6 +255,10 @@ class CORE_EXPORT QgsExpression
      * If the list contains a NULL QString, there is a variable name used
      * which is determined at runtime.
      *
+     * \note In contrast to the referencedColumns() function this method
+     * is not affected by any previous calls to QgsExpression::prepare(),
+     * or QgsExpressionNode::prepare().
+     *
      * \since QGIS 3.0
      */
     QSet<QString> referencedVariables() const;
@@ -256,6 +266,10 @@ class CORE_EXPORT QgsExpression
     /**
      * Returns a list of the names of all functions which are used in this expression.
      *
+     * \note In contrast to the referencedColumns() function this method
+     * is not affected by any previous calls to QgsExpression::prepare(),
+     * or QgsExpressionNode::prepare().
+     *
      * \since QGIS 3.2
      */
     QSet<QString> referencedFunctions() const;
@@ -294,6 +308,14 @@ class CORE_EXPORT QgsExpression
     /**
      * Returns a list of field name indexes obtained from the provided fields.
      *
+     * \warning If the expression has been prepared via a call to QgsExpression::prepare(),
+     * or a call to QgsExpressionNode::prepare() for a node has been made, then parts of
+     * the expression may have been determined to evaluate to a static pre-calculatable value.
+     * In this case the results will omit attribute indices which are used by these
+     * pre-calculated nodes, regardless of their actual referenced columns.
+     * If you are seeking to use these functions to introspect an expression you must
+     * take care to do this with an unprepared expression.
+     *
      * \since QGIS 3.0
      */
     QSet<int> referencedAttributeIndexes( const QgsFields &fields ) const;
@@ -327,10 +349,40 @@ class CORE_EXPORT QgsExpression
 
     /**
      * Checks whether an expression consists only of a single field reference
+     *
+     * \see expressionToLayerFieldIndex()
      * \since QGIS 2.9
      */
     bool isField() const;
 
+    /**
+     * Attempts to resolve an expression to a field index from the given \a layer.
+     *
+     * Given a string which may either directly match a field name from a layer, OR may
+     * be an expression which consists only of a single field reference for that layer, this
+     * method will return the corresponding field index.
+     *
+     * \returns field index if found, or -1 if \a expression does not represent a field from the layer.
+     *
+     * \see isField()
+     * \since QGIS 3.22
+     */
+    static int expressionToLayerFieldIndex( const QString &expression, const QgsVectorLayer *layer );
+
+    /**
+     * Validate if the expression is a field in the \a layer and ensure it is quoted.
+     *
+     * Given a string which may either directly match a field name from a layer, OR may
+     * be an expression which consists only of a single field reference for that layer, this
+     * method will return the quoted field.
+     *
+     * \returns the \a expression if not a field or quotes are not required, otherwise returns a quoted field.
+     *
+     * \see expressionToLayerFieldIndex()
+     * \since QGIS 3.24
+     */
+    static QString quoteFieldExpression( const QString &expression, const QgsVectorLayer *layer );
+
     /**
      * Tests whether a string is a valid expression.
      * \param text string to test
@@ -565,10 +617,25 @@ class CORE_EXPORT QgsExpression
      */
     static QStringList tags( const QString &name );
 
+    /**
+     * Adds a help string for a custom variable.
+     *
+     * The specified variable \a name should not have an existing help string set. If a help string is already present then
+     * FALSE will be returned and no changes will occur.
+     *
+     * \param name variable name
+     * \param description the help string to add. This is user visible, and should be a translated string.
+     * \returns TRUE if the help string was successfully added
+     * \see variableHelpText()
+     * \since QGIS 3.22
+     */
+    static bool addVariableHelpText( const QString name, const QString &description );
+
     /**
      * Returns the help text for a specified variable.
      * \param variableName name of variable
      * \see helpText()
+     * \see addVariableHelpText()
      * \since QGIS 2.12
      */
     static QString variableHelpText( const QString &variableName );
@@ -595,10 +662,11 @@ class CORE_EXPORT QgsExpression
      * length, and presents text representations of non numeric/text types (e.g., geometries and features).
      * \param value expression result to format
      * \param htmlOutput set to TRUE to allow HTML formatting, or FALSE for plain text output
+     * \param maximumPreviewLength define the maximum character length of the preview
      * \returns formatted string, may contain HTML formatting characters if \a htmlOutput is TRUE
      * \since QGIS 2.14
      */
-    static QString formatPreviewString( const QVariant &value, bool htmlOutput = true );
+    static QString formatPreviewString( const QVariant &value, bool htmlOutput = true, int maximumPreviewLength = 60 );
 
     /**
      * Create an expression allowing to evaluate if a field is equal to a
