@@ -147,13 +147,13 @@ void QgsRendererCategory::toSld( QDomDocument &doc, QDomElement &element, QVaria
   if ( mValue.isNull() || mValue.toString().isEmpty() )
   {
     filterFunc = QStringLiteral( "%1 = '%2' or %1 is null" )
-                 .arg( attrName.replace( '\"', QLatin1String( "\"\"" ) ),
+                 .arg( attrName.replace( '\"', QLatin1String( "\"\"" ) ).append( '"' ).prepend( '"' ),
                        mValue.toString().replace( '\'', QLatin1String( "''" ) ) );
   }
   else
   {
     filterFunc = QStringLiteral( "%1 = '%2'" )
-                 .arg( attrName.replace( '\"', QLatin1String( "\"\"" ) ),
+                 .arg( attrName.replace( '\"', QLatin1String( "\"\"" ) ).append( '"' ).prepend( '"' ),
                        mValue.toString().replace( '\'', QLatin1String( "''" ) ) );
   }
 
