@@ -42,7 +42,7 @@ QgsExtentWidget::QgsExtentWidget( QWidget *parent, WidgetStyle style )
   connect( mCondensedLineEdit, &QgsFilterLineEdit::cleared, this, &QgsExtentWidget::clear );
   connect( mCondensedLineEdit, &QLineEdit::textEdited, this, &QgsExtentWidget::setOutputExtentFromCondensedLineEdit );
 
-  mLayerMenu = new QMenu( tr( "Calculate from Layer" ) );
+  mLayerMenu = new QMenu( tr( "Calculate from Layer" ), this );
   mButtonCalcFromLayer->setMenu( mLayerMenu );
   connect( mLayerMenu, &QMenu::aboutToShow, this, &QgsExtentWidget::layerMenuAboutToShow );
   mMapLayerModel = new QgsMapLayerModel( this );
@@ -123,7 +123,7 @@ void QgsExtentWidget::setOutputCrs( const QgsCoordinateReferenceSystem &outputCr
   mHasFixedOutputCrs = true;
   if ( mOutputCrs != outputCrs )
   {
-    bool prevExtentEnabled = mIsValid;
+    const bool prevExtentEnabled = mIsValid;
     switch ( mExtentState )
     {
       case CurrentExtent:
@@ -150,7 +150,8 @@ void QgsExtentWidget::setOutputCrs( const QgsCoordinateReferenceSystem &outputCr
         try
         {
           QgsCoordinateTransform ct( mOutputCrs, outputCrs, QgsProject::instance() );
-          QgsRectangle extent = ct.transformBoundingBox( outputExtent() );
+          ct.setBallparkTransformsAreAppropriate( true );
+          const QgsRectangle extent = ct.transformBoundingBox( outputExtent() );
           mOutputCrs = outputCrs;
           setOutputExtentFromUser( extent, outputCrs );
         }
@@ -186,6 +187,7 @@ void QgsExtentWidget::setOutputExtent( const QgsRectangle &r, const QgsCoordinat
       try
       {
         QgsCoordinateTransform ct( srcCrs, mOutputCrs, QgsProject::instance() );
+        ct.setBallparkTransformsAreAppropriate( true );
         extent = ct.transformBoundingBox( r );
       }
       catch ( QgsCsException & )
@@ -252,10 +254,11 @@ void QgsExtentWidget::setOutputExtentFromCondensedLineEdit()
     const QRegularExpressionMatch match = mCondensedRe.match( text );
     if ( match.hasMatch() )
     {
-      whileBlocking( mXMinLineEdit )->setText( match.captured( 1 ) );
-      whileBlocking( mXMaxLineEdit )->setText( match.captured( 2 ) );
-      whileBlocking( mYMinLineEdit )->setText( match.captured( 3 ) );
-      whileBlocking( mYMaxLineEdit )->setText( match.captured( 4 ) );
+      // Localization
+      whileBlocking( mXMinLineEdit )->setText( QLocale().toString( match.captured( 1 ).toDouble() ) );
+      whileBlocking( mXMaxLineEdit )->setText( QLocale().toString( match.captured( 2 ).toDouble() ) );
+      whileBlocking( mYMinLineEdit )->setText( QLocale().toString( match.captured( 3 ).toDouble() ) );
+      whileBlocking( mYMaxLineEdit )->setText( QLocale().toString( match.captured( 4 ).toDouble() ) );
       if ( !match.captured( 5 ).isEmpty() )
       {
         mOutputCrs = QgsCoordinateReferenceSystem( match.captured( 5 ) );
@@ -270,7 +273,7 @@ void QgsExtentWidget::setOutputExtentFromCondensedLineEdit()
 
 void QgsExtentWidget::clear()
 {
-  bool prevWasNull = mIsValid;
+  const bool prevWasNull = mIsValid;
 
   whileBlocking( mXMinLineEdit )->clear();
   whileBlocking( mXMaxLineEdit )->clear();
@@ -315,12 +318,12 @@ void QgsExtentWidget::layerMenuAboutToShow()
   mLayerMenu->clear();
   for ( int i = 0; i < mMapLayerModel->rowCount(); ++i )
   {
-    QModelIndex index = mMapLayerModel->index( i, 0 );
-    QIcon icon = qvariant_cast<QIcon>( mMapLayerModel->data( index, Qt::DecorationRole ) );
-    QString text = mMapLayerModel->data( index, Qt::DisplayRole ).toString();
+    const QModelIndex index = mMapLayerModel->index( i, 0 );
+    const QIcon icon = qvariant_cast<QIcon>( mMapLayerModel->data( index, Qt::DecorationRole ) );
+    const QString text = mMapLayerModel->data( index, Qt::DisplayRole ).toString();
     QAction *act = new QAction( icon, text, mLayerMenu );
     act->setToolTip( mMapLayerModel->data( index, Qt::ToolTipRole ).toString() );
-    QString layerId = mMapLayerModel->data( index, QgsMapLayerModel::LayerIdRole ).toString();
+    const QString layerId = mMapLayerModel->data( index, QgsMapLayerModel::LayerIdRole ).toString();
     if ( mExtentState == ProjectLayerExtent && mExtentLayer && mExtentLayer->id() == layerId )
     {
       act->setCheckable( true );
@@ -433,16 +436,17 @@ QgsRectangle QgsExtentWidget::outputExtent() const
                        QgsDoubleValidator::toDouble( mXMaxLineEdit->text() ), QgsDoubleValidator::toDouble( mYMaxLineEdit->text() ) );
 }
 
-void QgsExtentWidget::setMapCanvas( QgsMapCanvas *canvas )
+void QgsExtentWidget::setMapCanvas( QgsMapCanvas *canvas, bool drawOnCanvasOption )
 {
   if ( canvas )
   {
     mCanvas = canvas;
-    mButtonDrawOnCanvas->setVisible( true );
+    mButtonDrawOnCanvas->setVisible( drawOnCanvasOption );
     mCurrentExtentButton->setVisible( true );
 
     mMenu->addAction( mUseCanvasExtentAction );
-    mMenu->addAction( mDrawOnCanvasAction );
+    if ( drawOnCanvasOption )
+      mMenu->addAction( mDrawOnCanvasAction );
   }
   else
   {
