@@ -35,6 +35,7 @@
 #include <algorithm>
 #include <QDir>
 #include <QTimer>
+#include <QUrlQuery>
 
 QgsWFSFeatureHitsAsyncRequest::QgsWFSFeatureHitsAsyncRequest( QgsWFSDataSourceURI &uri )
   : QgsWfsRequest( uri )
@@ -106,7 +107,7 @@ QString QgsWFSFeatureDownloaderImpl::sanitizeFilter( QString filter )
   return filter;
 }
 
-QUrl QgsWFSFeatureDownloaderImpl::buildURL( qint64 startIndex, int maxFeatures, bool forHits )
+QUrl QgsWFSFeatureDownloaderImpl::buildURL( qint64 startIndex, long long maxFeatures, bool forHits )
 {
   QUrl getFeatureUrl( mShared->mURI.requestUrl( QStringLiteral( "GetFeature" ) ) );
   QUrlQuery query( getFeatureUrl );
@@ -122,7 +123,7 @@ QUrl QgsWFSFeatureDownloaderImpl::buildURL( qint64 startIndex, int maxFeatures,
   else
   {
     QSet<QString> setNamespaces;
-    for ( const QgsOgcUtils::LayerProperties &layerProperties : qgis::as_const( mShared->mLayerPropertiesList ) )
+    for ( const QgsOgcUtils::LayerProperties &layerProperties : std::as_const( mShared->mLayerPropertiesList ) )
     {
       if ( !typenames.isEmpty() )
         typenames += QLatin1Char( ',' );
@@ -138,8 +139,13 @@ QUrl QgsWFSFeatureDownloaderImpl::buildURL( qint64 startIndex, int maxFeatures,
     }
   }
   if ( mShared->mWFSVersion.startsWith( QLatin1String( "2.0" ) ) )
+  {
     query.addQueryItem( QStringLiteral( "TYPENAMES" ),  typenames );
-  query.addQueryItem( QStringLiteral( "TYPENAME" ),  typenames );
+  }
+  else
+  {
+    query.addQueryItem( QStringLiteral( "TYPENAME" ),  typenames );
+  }
 
   if ( forHits )
   {
@@ -205,10 +211,12 @@ QUrl QgsWFSFeatureDownloaderImpl::buildURL( qint64 startIndex, int maxFeatures,
     QString geometryAttribute( mShared->mGeometryAttribute );
     if ( mShared->mLayerPropertiesList.size() > 1 )
       geometryAttribute = mShared->mURI.typeName() + "/" + geometryAttribute;
-    else if ( mShared->mLayerPropertiesList.size() == 1 && !mShared->mLayerPropertiesList[0].mNamespacePrefix.isEmpty() )
-      geometryAttribute = mShared->mLayerPropertiesList[0].mNamespacePrefix + QStringLiteral( ":" ) + geometryAttribute;
+
     QDomElement bboxElem = QgsOgcUtils::expressionToOgcFilter( bboxExp, doc,
-                           gmlVersion, filterVersion, geometryAttribute, mShared->srsName(),
+                           gmlVersion, filterVersion,
+                           mShared->mLayerPropertiesList.size() == 1 ? mShared->mLayerPropertiesList[0].mNamespacePrefix : QString(),
+                           mShared->mLayerPropertiesList.size() == 1 ? mShared->mLayerPropertiesList[0].mNamespaceURI : QString(),
+                           geometryAttribute, mShared->srsName(),
                            honourAxisOrientation, mShared->mURI.invertAxisOrientation() );
     doc.appendChild( bboxElem );
     QDomNode bboxNode = bboxElem.firstChildElement();
@@ -224,16 +232,25 @@ QUrl QgsWFSFeatureDownloaderImpl::buildURL( qint64 startIndex, int maxFeatures,
     andElem.appendChild( filterNode );
     doc.firstChildElement().appendChild( andElem );
 
-    QSet<QString> setNamespaceURI;
-    for ( const QgsOgcUtils::LayerProperties &props : qgis::as_const( mShared->mLayerPropertiesList ) )
+    if ( mShared->mLayerPropertiesList.size() == 1 &&
+         doc.firstChildElement().hasAttribute( QStringLiteral( "xmlns:" ) + mShared->mLayerPropertiesList[0].mNamespacePrefix ) )
     {
-      if ( !props.mNamespacePrefix.isEmpty() && !props.mNamespaceURI.isEmpty() &&
-           !setNamespaceURI.contains( props.mNamespaceURI ) )
+      // nothing to do
+    }
+    else
+    {
+      // add xmls:PREFIX=URI attributes to top element
+      QSet<QString> setNamespaceURI;
+      for ( const QgsOgcUtils::LayerProperties &props : std::as_const( mShared->mLayerPropertiesList ) )
       {
-        setNamespaceURI.insert( props.mNamespaceURI );
-        QDomAttr attr = doc.createAttribute( QStringLiteral( "xmlns:" ) + props.mNamespacePrefix );
-        attr.setValue( props.mNamespaceURI );
-        doc.firstChildElement().setAttributeNode( attr );
+        if ( !props.mNamespacePrefix.isEmpty() && !props.mNamespaceURI.isEmpty() &&
+             !setNamespaceURI.contains( props.mNamespaceURI ) )
+        {
+          setNamespaceURI.insert( props.mNamespaceURI );
+          QDomAttr attr = doc.createAttribute( QStringLiteral( "xmlns:" ) + props.mNamespacePrefix );
+          attr.setValue( props.mNamespaceURI );
+          doc.firstChildElement().setAttributeNode( attr );
+        }
       }
     }
 
@@ -367,7 +384,7 @@ void QgsWFSFeatureDownloaderImpl::createProgressDialog()
   CONNECT_PROGRESS_DIALOG( QgsWFSFeatureDownloaderImpl );
 }
 
-void QgsWFSFeatureDownloaderImpl::run( bool serializeFeatures, int maxFeatures )
+void QgsWFSFeatureDownloaderImpl::run( bool serializeFeatures, long long maxFeatures )
 {
   bool success = true;
 
@@ -423,9 +440,7 @@ void QgsWFSFeatureDownloaderImpl::run( bool serializeFeatures, int maxFeatures )
     {
       break;
     }
-    int maxFeaturesThisRequest = static_cast<int>(
-                                   std::min( maxTotalFeatures - mTotalDownloadedFeatureCount,
-                                       static_cast<qint64>( std::numeric_limits<int>::max() ) ) );
+    long long maxFeaturesThisRequest = maxTotalFeatures - mTotalDownloadedFeatureCount;
     if ( mShared->mPageSize > 0 )
     {
       if ( maxFeaturesThisRequest > 0 )
@@ -454,7 +469,7 @@ void QgsWFSFeatureDownloaderImpl::run( bool serializeFeatures, int maxFeatures )
              true, /* forceRefresh */
              false /* cache */ );
 
-    int featureCountForThisResponse = 0;
+    long long featureCountForThisResponse = 0;
     bool bytesStillAvailableInReply = false;
     // Loop until there is no data coming from the current request
     while ( true )
