@@ -11,7 +11,7 @@ __author__ = 'Alessandro Pasotti'
 __date__ = '22/01/2020'
 __copyright__ = 'Copyright 2020, The QGIS Project'
 # This will get replaced with a git SHA1 when you do a git archive
-__revision__ = '4279198632348b51b6985e3c3012256e953c411d'
+__revision__ = '6f08e4d7b004dfd6b584a5c2ac195a6470cbb531'
 
 import tempfile
 from qgis.core import (
@@ -44,7 +44,7 @@ class TestPyQgsVectorLayerNamedStyle(unittest.TestCase):
         self.assertTrue(vl.writeStyle(node, style, "Error writing style", QgsReadWriteContext(), QgsMapLayer.Rendering))
 
         style_content = style.toString()
-        del(vl)
+        del vl
 
         # Read
         vl2 = QgsVectorLayer("LineString?crs=epsg:4326", "result", "memory")
